#' VOSON Dashboard Text Analysis Module
#'
#' Shiny module functions to create and render groups of word frequency charts or 
#' word cloud plots dynamically.
#'

#' Plot variables
#' todo: pass these in some other way
tamod_palette <- g_plot_palette # brewer.pal(8, "Dark2")
tamod_plot_height <- "450px"
tamod_wc_scale <- c(2, 0.5)

#' Create empty plot with text message.
#' 
#' @param message text message to print in centre of empty plot
#' 
#' @return plot and text code block
#' 
emptyPlotMessage <- function(message) {
  return({ plot(1:10, 1:10, type = "n", axes = F, xlab = "", ylab = "")
           text(5, 5, message, cex = 1.2) })
}

#' UI container for Text Analysis plots.
#' Creates namespaced UI output container \code{"ta_plot"}.
#' 
#' @param id shiny module namespace id
#' 
#' @return None
#' 
#' @export
taPlotContainerUI <- function(id) {
  ns <- NS(id)
  uiOutput(ns("ta_plot"))
}

#' Creates UI plot placeholders generated from VOSON Dashboard Text Analysis plot list data.
#' Renders plot placeholders in namespace UI output container \code{"ta_plot"}.
#' 
#' @param input shiny module namespaced \code{"input"} parameter
#' @param output shiny module namespaced \code{"output"} parameter
#' @param session shiny module namespaced \code{"session"} parameter
#' @param data voson dashboard text analysis plot list data structure
#'    data structure is named list: \code{"data[plot-id] <- list(graph_attr = list(attribute_name, attribute_value), VCorpus)"}
#'
#' @return None
#' 
#' @export
taPlotPlaceholders <- function(input, output, session, data) {
  ns = session$ns
  
  plotPlaceholders <- reactive({
    tag_list <- tagList()
    plot_ids <- names(sapply(data, names))
    
    # isolate({ # withProgress(message = "Processing placeholders...", {
      
      for (i in seq_along(data)) {
        title_cat <- data[[i]][[1]][[1]]
        title_attr <- paste0(data[[i]][[1]][[2]], collapse = " / ")
        
        title_tags <- fluidRow(column(width = 12, div(h4(title_cat, " ", title_attr), style = "padding-left:20px;")))
        tag_list <- tagAppendChild(tag_list, title_tags)
        
        plot_id <- ns(plot_ids[i])
        plot_tags <- fluidRow(column(width = 12, plotOutput(plot_id, height = tamod_plot_height)))
        tag_list <- tagAppendChild(tag_list, plot_tags)
      }
      
    # }) # })

    # creates a plot placeholder with the namespace id "no-data"
    if (length(tag_list) == 0) {
      tag_list <- tagAppendChild(tag_list, fluidRow(column(width = 12, plotOutput(ns("no-data"), height = tamod_plot_height))))
    }
    
    return(tag_list)
  })
  
  output$ta_plot <- renderUI({
    plotPlaceholders()
  })
}

#' Renders plots to placeholders for Word Frequency or Word Cloud plots.
#' Generated from VOSON Dashboard Text Analysis plot list data structure.
#'
#' @param input shiny module namespaced \code{"input"} parameter
#' @param output shiny module namespaced \code{"output"} parameter
#' @param session shiny module namespaced \code{"session"} parameter
#' @param data voson dashboard text analysis plot list data structure \cr
#'    data structure is named list: \code{"data[plot-id] <- list(graph_attr = list(attribute_name, attribute_value), VCorpus)"}
#' @param seed value to seed rendering of word clouds
#' @param categories list of all categorical attribures in the data set
#' @param min_freq minimum word frequency for word frequency charts or word clouds
#' @param max_words maximum number of words to render in word clouds
#' @param top_count number of words to render in word frequency charts
#' @param type code for type of plots \code{"wf"} word frequencies or \code{"wc"} word clouds
#' 
#' @return None
#'
#' @export
taPlotList <- function(input, output, session, data, seed, categories, min_freq, max_words, top_count, type) {
  ns = session$ns
  
  wordFreqPlotList <- reactive({
    plot_ids <- names(sapply(data, names))
    
    isolate({ withProgress(message = "Processing frequency plots...", {
      
      for (i in seq_along(data)) {
        local({
          local_i <- i
          plot_id <- plot_ids[local_i]
          output[[plot_id]] <- renderPlot({
            wordFreqChart(data[[local_i]], categories, min_freq, top_count)
          })
        })
      }
  
      # renders empty plot to namespace id "no-data" placeholder
      if (length(data) < 1) {
        output[["no-data"]] <- renderPlot({
          emptyPlotMessage("No text data.")
        })
      }
      
    }) })
  })
  
  wordCloudPlotList <- reactive({
    plot_ids <- names(sapply(data, names))
    
    isolate({ withProgress(message = "Processing cloud plots...", {
      
      for (i in seq_along(data)) {
        local({
          local_i <- i
          plot_id <- plot_ids[local_i]
          output[[plot_id]] <- renderPlot({
            wordCloudPlot(data[[local_i]], seed, categories, min_freq, max_words)
          })
        })
      }
      
      # renders empty plot to namespace id "no-data" placeholder
      if (length(data) < 1) {
        output[["no-data"]] <- renderPlot({
          emptyPlotMessage("No text data.")
        })
      }
      
    }) })
  })
  
  # determine type of plots to generate
  if (type == "wf") {
    wordFreqPlotList()
  } else if (type == "wc") {
    wordCloudPlotList()
  }
}

#' Creates a Word Frequency chart
#'
#' @param data voson dashboard text analysis plot list data structure
#' @param categories list of all categorical attribures in the data set
#' @param min_freq minimum word frequency for word frequency charts
#' @param top_count number of words to render in word frequency charts
#' 
#' @return barchart
#' 
wordFreqChart <- function(data, categories, min_freq, top_count) {
  graph_attr <- data[[1]]
  corp <- data[[2]]
  
  plot_category <- plot_category_attr <- ""
  
  if (missing(graph_attr)) {
    graph_attr <- NULL
  } else {
    plot_category <- graph_attr[[1]]
    plot_category_attr <- graph_attr[[2]]
  }
  
  # min freq uses bounds control
  dtm <- DocumentTermMatrix(corp, control = list(wordLengths = c(3, 20), bounds = list(global = c(min_freq, Inf))))
  dtmSparseRemoved <- removeSparseTerms(dtm, 0.98)
  
  freqTerms <- colSums(as.matrix(dtmSparseRemoved))
  orderTerms <- order(freqTerms, decreasing = TRUE)
  
  if (plot_category != "") {
    df <- data.frame('cat' = categories[[plot_category]])
    df$color <- tamod_palette[1:nrow(df)]
    match_t <- match(plot_category_attr, df$cat)
    colx <- ifelse(!is.na(match_t), df$color[match_t], 'white')
  } else {
    colx <- 'white'
  }
  
  # returns empty plot with message if no data to chart
  if (is.null(corp) || length(corp) < 1) {
    return(emptyPlotMessage("No text data."))
  }
  par(mar = rep(0, 4))
  return(barchart(freqTerms[orderTerms[1:top_count]], col = colx, xlab = "Frequency"))
}

#' Creates a Word Cloud plot
#'
#' @param data voson dashboard text analysis plot list data structure
#' @param seed value to seed rendering of word clouds
#' @param categories list of all categorical attribures in the data set
#' @param min_freq minimum word frequency for word clouds
#' @param max_words maximum number of words to render in word clouds
#' 
#' @return wordcloud
#' 
wordCloudPlot <- function(data, seed, categories, min_freq, max_words) {
  graph_attr <- data[[1]]
  corp <- data[[2]]
  
  plot_category <- plot_category_attr <- ""
  
  if (missing(graph_attr)) {
    graph_attr <- NULL
  } else {
    plot_category <- graph_attr[[1]]
    plot_category_attr <- graph_attr[[2]]
  }
  
  if (plot_category != "") {
    df <- data.frame('cat' = categories[[plot_category]])
    df$color <- tamod_palette[1:nrow(df)]
    match_t <- match(plot_category_attr, df$cat)
    colx <- ifelse(!is.na(match_t), df$color[match_t], 'black')
  } else {
    colx <- 'black'
  }
  
  # returns empty plot with message if no data to plot
  if (is.null(corp) || length(corp) < 1) {
    return(emptyPlotMessage("No text data."))
  }
  
  if (!is.null(seed)) {
    set.seed(seed)
  }
  par(mar = rep(0, 4))
  return(wordcloud(corp, min.freq = min_freq, max.words = max_words, random.order = FALSE, colors = colx))
}
